/* $Id: dr2proxy.c,v 1.5 2002/04/16 12:42:39 d Exp $ */
/* David Leonard, 2002. Public domain. */
/*
 * Provide inter-LAN proxy service for Dark Reign 2.
 * DR2 solicits LAN games by sending UDP packets to udp address *:0x6666
 * containing 08 80 37 55.
 * A DR2 game responds with a UDP packet containing 0x38 bytes containing
 * the name of the game.
 * The client game then connects to tcp port 0x6666.
 *
 *	S----A----B---C
 * This program runs as one half of a symmetric peer bridge.
 * - A listens for broadcasts messages from S
 * - A forwards the messages (encapsulated) to B, who broadcasts them to C
 *   but forged to appear to come directly from S
 * - C begins directly communicating with S
 *
 * 
 */

#include <err.h>
#include <errno.h>
#include <netdb.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ifaddrs.h>

#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>
#ifdef USE_KQ
#include <sys/event.h>		/* kevent syscall */
#else
#include <poll.h>		/* poll syscall */
#endif

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/udp.h>

#include <net/if.h>

#include <arpa/inet.h>

#define VERSION		1.1

#define DEFBPORT	26214		/* Dark Reign 2 port */
#define DEFLPORT	26213		/* inter-proxy port */
#define MAXREMOTE	32		/* max # of Dark Reign 2 clients */
#define MAXPEERS	32		/* max # of inter-proxy peers */

struct buf {
	struct in_addr	addr;
	in_port_t	port;
	char		data[32768];
};
#define offsetof(t,f) ((int)&((t*)0)->f)

int debug = 0;
int local_count = 0;
int remote_count = 0;
volatile int got_siginfo;

/* Handler for SIGINFO signals */
void
siginfo(sig)
	int sig;
{
	got_siginfo = 1;
}

/* Create a raw socket */
int
rawsocket(bindaddr, bindlen)
	struct sockaddr *bindaddr;
	socklen_t bindlen;
{
	int r, flag;

	/* Allocate a raw packet socket */
	if ((r = socket(AF_INET, SOCK_RAW, 0)) == -1)
		err(1, "socket");

	/* Bind the raw socket to that interface */
	if (debug)
		fprintf(stderr, "binding to interface at %s\n",
		    inet_ntoa(((struct sockaddr_in *)bindaddr)->sin_addr));
	if (bind(r, bindaddr, bindlen) == -1)
		err(1, "bind");

	/* Set the raw socket to header-include mode */
	flag = 1;
	if (setsockopt(r, IPPROTO_IP, IP_HDRINCL, &flag, sizeof(flag)) == -1)
		err(1, "setsockopt");

	return r;
}

/* Send a UDP packet on a raw socket */
int
sendudp(r, src, dst, data, len)
	int r;
	struct sockaddr_in *src, *dst;
	char *data;
	size_t len;
{
	char buf[len + sizeof (struct ip) + sizeof (struct udphdr)];
	char *p;
	struct ip *ip;
	struct udphdr *udp;
	int i;

	p = buf;

	/* IPv4 header */
	ip = (struct ip *)p;
	ip->ip_v = IPVERSION;
	ip->ip_hl = sizeof (struct ip) >> 2;
	ip->ip_tos = 0;
	ip->ip_id = 0;		/* kernel gives packet an ID */
	ip->ip_off = 0;
	ip->ip_ttl = IPDEFTTL;
	ip->ip_p = IPPROTO_UDP;
	ip->ip_sum = 0;
	ip->ip_src.s_addr = src->sin_addr.s_addr;
	ip->ip_dst.s_addr = dst->sin_addr.s_addr;
	p = p + sizeof (struct ip);

	/* UDPv4 header */
	udp = (struct udphdr *)p;
	udp->uh_sport = src->sin_port;
	udp->uh_dport = dst->sin_port;
	p = p + sizeof (struct udphdr);

	/* UDP payload */
	memcpy(p, data, len);
	p += len;

	udp->uh_ulen = htons(p - (char *)udp);
	ip->ip_len = htons(p - buf);

	/* XXX should compute UDP checksum here */
	udp->uh_sum = 0;		/* don't care (RFC768) */

	if (debug) {
		fprintf(stderr, "sending raw packet:");
		for (i = 0; i < (p-buf); i++) {
			if (i % 0x10 == 0) fprintf(stderr, "\n  %04x:", i);
			if (i % 2 == 0) fprintf(stderr, " ");
			fprintf(stderr, "%02x", (unsigned char)buf[i]);
		}
		fprintf(stderr, "\n");
	}

	return sendto(r, buf, p - buf, 0, (struct sockaddr *)dst, sizeof *dst);
}

int
main(argc, argv)
	int argc;
	char *argv[];
{
	struct sockaddr_in here, bcast;
	struct {
		struct sockaddr_in addr;
		socklen_t len;
	} peer[MAXPEERS];
	int npeers = 0;
	int r, l, d;
	char *localif = "lo0";
	int ch;
	extern char *optarg;
	extern int optind;
	struct ifaddrs *ifa, *ifp;
	int error = 0;
	int lport = DEFLPORT;
	char *broadip = NULL;
	struct addrinfo hints;
#ifdef USE_KQ
	int kq;
	struct kevent ke[2], e;
#else
	struct pollfd pfd[2];
#endif
	struct in_addr remote[MAXREMOTE];
	int nremote = 0;
	int bport = DEFBPORT;

	/* Parse arguments */
	while ((ch = getopt(argc, argv, "di:l:b:P:")) != -1)
	switch (ch) {
		case 'd':
			debug = 1;
			break;
		case 'i':
			localif = optarg;
			break;
		case 'l':
			lport = atoi(optarg);
			break;
		case 'b':
			broadip = optarg;
			break;
		case 'P':
			bport = atoi(optarg);
			break;
		default:
			error = 1;
			break;
	}
	if (argc - optind == 0)
		error = 1;
	if (error) {
		fprintf(stderr, "dr2proxy version %f\n", VERSION);
		fprintf(stderr, "usage: %s "
		    "[-d] [-i interface] [-l port] [-b bcastip] "
		    "[-P gameport] host[:port] ...\n",
		    argv[0]);
		exit(1);
	}

	/* Listen to the DR2 UDP broadcast port */
	l = socket(PF_INET, SOCK_DGRAM, 0);
	if (l == -1)
		err(1, "socket");
	memset(&here, 0, sizeof here);
	here.sin_family = AF_INET;
	here.sin_addr.s_addr = INADDR_ANY;
	here.sin_port = htons(bport);
	if (bind(l, (struct sockaddr *)&here, sizeof here) == -1)
		err(1, "bind");

	/* Get the addresses of the peer proxies */
	memset (&hints, 0, sizeof hints);
	hints.ai_family = PF_INET;
	hints.ai_socktype = SOCK_DGRAM;
	if ((d = socket(hints.ai_family, hints.ai_socktype, 0)) == -1)
		err(1, "socket");
	for (; optind < argc; optind ++) {
	    struct addrinfo *res, *res0;
	    char defport[16];
	    char *host = argv[optind];
	    char *port = NULL;

	    port = strrchr(host, ':');
	    if (!port) {
		snprintf(defport, sizeof defport, "%u", lport);
		port = defport;
	    } else {
		*port++ = '\0';
	    }
	    error = getaddrinfo(host, port, &hints, &res0);
	    if (error)
		errx(1, "%s: %s:%s", host, port, gai_strerror(error));
	    for (res = res0; res; res = res->ai_next) {
		if (res->ai_family != PF_INET || 
		    res->ai_socktype != SOCK_DGRAM)
			continue;
		if (npeers >= MAXPEERS)
			errx(1, "too many peers");
		peer[npeers].len = res->ai_addrlen;
		if (res->ai_addrlen > sizeof peer[npeers].addr)
			errx(1, "ai_addrlen > sizeof addr");
		memcpy(&peer[npeers].addr, res->ai_addr, res->ai_addrlen);
		if (debug)
		    fprintf(stderr, "using peer at %s:%u\n", 
		        inet_ntoa(peer[npeers].addr.sin_addr),
		        ntohs(peer[npeers].addr.sin_port));
		npeers++;
		break;
	    }
	    if (!res)
		errx(1, "%s:%s: no address", host, port);
	    freeaddrinfo(res0);
	}

	memset(&here, 0, sizeof here);
	here.sin_family = AF_INET;
	here.sin_addr.s_addr = INADDR_ANY;
	here.sin_port = htons(lport);
	if (bind(d, (struct sockaddr *)&here, sizeof here) == -1) 
		err(1, "bind");

	/* Construct raw socket on an interface */
	if (getifaddrs(&ifa) == -1)
		err(1, "getifaddrs");
	for (ifp = ifa; ifp; ifp = ifp->ifa_next) 
		if (strcmp(ifp->ifa_name, localif) == 0 &&
		    (ifp->ifa_flags & IFF_UP) == IFF_UP &&
		    ifp->ifa_addr->sa_family == AF_INET &&
		    ifp->ifa_broadaddr != NULL)
			break;
	if (!ifp)
		errx(1, "%s not found or not AF_INET", localif);
	bcast = *((struct sockaddr_in *)ifp->ifa_broadaddr);
	r = rawsocket(ifp->ifa_addr, ifp->ifa_addr->sa_len);
	freeifaddrs(ifa);
	bcast.sin_port = htons(bport);

	/* Allow override of broadcast address */
	if (broadip) {
		if (!inet_aton(broadip, &bcast.sin_addr))
			errx(1, "invalid IP address: %s", broadip);
		if (debug)
			fprintf(stderr, "using broadcast address %s\n",
			    broadip);
	}

#if defined(SIGINFO)
	if (signal(SIGINFO, siginfo) == SIG_ERR)
		warn("signal");
#endif

	/* Wait for a UDP packet from client or peer */
#ifdef USE_KQ
	if ((kq = kqueue()) == -1)
		err(1, "kqueue");
	EV_SET(&ke[0], l, EVFILT_READ, EV_ADD, 0, 0, NULL);
	EV_SET(&ke[1], d, EVFILT_READ, EV_ADD, 0, 0, NULL);
	if (kevent(kq, ke, 2, NULL, 0, NULL) == -1)
		err(1, "kevent");
#endif
	if (debug)
		fprintf(stderr, "waiting for packets...\n");
	for (;;) {

#ifdef USE_KQ
		if ((error = kevent(kq, NULL, 0, &e, 1, NULL)) == -1
		    && errno != EINTR)
			err(1, "kevent");
#else
		pfd[0].fd = l;
		pfd[0].events = POLLIN;
		pfd[0].revents = 0;
		pfd[1].fd = d;
		pfd[1].events = POLLIN;
		pfd[1].revents = 0;
		if ((error = poll(pfd, 2, -1)) == -1 && errno != EINTR)
			err(1, "poll");
#endif
		if (got_siginfo) {
			warnx("broadcasts heard: "
			    "%d local, %d remote\n", local_count,
			    remote_count);
			got_siginfo = 0;
		}

		if (error == -1)
			continue;

		if (debug)
			fprintf(stderr, "got something\n");
#if USE_KQ
		if (e.ident == l) 
#else
		if (pfd[0].revents)
#endif
		{
			/* Received packet from a local DR2 game */
			struct buf buf;
			int len, i;
			struct sockaddr_in from;
			socklen_t fromlen = sizeof from;

			len = recvfrom(l, buf.data, sizeof buf.data, 0,
			    (struct sockaddr *)&from, &fromlen);
			if (len == -1)
				err(1, "recvfrom bcast");

			local_count++;

			if (debug) {
				fprintf(stderr, "heard %d bcast from %s:%u\n",
				    len, inet_ntoa(from.sin_addr),
				    ntohs(from.sin_port));
			}
			/* Have we seen this packet from a peer host? */
			for (i = 0; i < nremote; i++)
				if (memcmp(&remote[i], &from.sin_addr,
				    sizeof from.sin_addr) == 0)
					break;
			if (i < nremote) {
				if (debug) 
					fprintf(stderr, "  - ignoring\n");
				continue;
			} 
			/* Fill in the rest of the encap'd packet */
		        buf.addr = from.sin_addr;
		        buf.port = from.sin_port;
			/* Send it to each of the given peers */
			for (i = 0; i < npeers; i++) {
			    if (debug) {
				fprintf(stderr, "  - sending to %s:%u\n",
				    inet_ntoa(peer[i].addr.sin_addr),
				    ntohs(peer[i].addr.sin_port));
			    }
			    if (sendto(d, &buf, len +
				offsetof(struct buf, data),
			        0, (struct sockaddr *)&(peer[i].addr),
				peer[i].len) == -1)
				    warn("sendto");
			}
		}
		else
#ifdef USE_KQ
		if (e.ident == d) 
#else
		if (pfd[1].revents)
#endif
		{
			/* Received packet from peer */
			struct buf buf;
			int len, i;
			struct sockaddr_in from;
			socklen_t flen = sizeof from;

			len = recvfrom(d, &buf, sizeof buf, 0, 
			    (struct sockaddr *)&from, &flen);
			if (len == -1)
				err(1, "recvfrom peer");

			remote_count++;

			if (debug)
				fprintf(stderr, "forging %d bcast from %s:%u\n",
				    len - offsetof(struct buf, data), 
				    inet_ntoa(buf.addr), ntohs(buf.port));

			memset(&from, 0, sizeof from);
			from.sin_family = AF_INET;
			from.sin_port = buf.port;
			from.sin_addr = buf.addr;
			/* Send it out as a forged packet */
			if (sendudp(r, &from, &bcast, &buf.data, 
			    len - offsetof(struct buf, data)) == -1)
				warn("sendudp");

			/*
			 * Record it in the remote table so we ignore
			 * it when it comes back on the broadcast interface.
			 */
			for (i = 0; i < nremote; i++)
				if (memcmp(&remote[i], &from.sin_addr,
				    sizeof from.sin_addr) == 0)
					break;
			if (i >= nremote) {
				if (i >= MAXREMOTE)
					warnx("remote hosts list full");
				else
					memcpy(&remote[nremote++], 
					    &from.sin_addr,
					    sizeof from.sin_addr);
			}
		}
	}

	exit(0);
}
